import argparse
import time
from os.path import basename, join

import imageio
import numpy as np

import pyastar2d


def parse_args():
    parser = argparse.ArgumentParser('An example of using pyastar2d to find the solution to a maze')
    parser.add_argument(
        '--input',
        type=str,
        default='mazes/maze_small.png',
        help='Path to the black-and-white image to be used as input.',
    )
    parser.add_argument(
        '--output',
        type=str,
        help='Path to where the output will be written',
    )

    args = parser.parse_args()

    if args.output is None:
        args.output = join('solns', basename(args.input))

    return args


def main():
    args = parse_args()
    maze = imageio.v3.imread(args.input, mode='L')

    if maze is None:
        print(f'No file found: {args.input}')
        return
    else:
        print(f'Loaded maze of shape {maze.shape} from {args.input}')

    grid = maze.astype(np.float32)
    grid[grid == 0] = np.inf
    grid[grid == 255] = 1

    assert grid.min() == 1, 'cost of moving must be at least 1'

    # start is the first white block in the top row
    (start_j,) = np.where(grid[0, :] == 1)
    start = np.array([0, start_j[0]])

    # end is the first white block in the final column
    (end_i,) = np.where(grid[:, -1] == 1)
    end = np.array([end_i[0], grid.shape[0] - 1])

    t0 = time.time()
    # set allow_diagonal=True to enable 8-connectivity
    path = pyastar2d.astar_path(grid, start, end, allow_diagonal=False)
    dur = time.time() - t0

    if path.shape[0] > 0:
        print(f'Found path of length {path.shape[0]} in {dur:.6f}s')
        maze = np.stack((maze, maze, maze), axis=2)
        maze[path[:, 0], path[:, 1]] = (255, 0, 0)

        print(f'Plotting path to {args.output}')
        imageio.imwrite(args.output, maze)
    else:
        print('No path found')

    print('Done')


if __name__ == '__main__':
    main()



import numpy as np
import pytest

import pyastar2d
from pyastar2d import Heuristic


def test_small():
    weights = np.array(
        [
            [1, 3, 3, 3, 3],
            [2, 1, 3, 3, 3],
            [2, 2, 1, 3, 3],
            [2, 2, 2, 1, 3],
            [2, 2, 2, 2, 1],
        ],
        dtype=np.float32,
    )
    # Run down the diagonal.
    path = pyastar2d.astar_path(weights, (0, 0), (4, 4), allow_diagonal=True)
    expected = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [4, 4]])

    assert np.all(path == expected)

    # Down, right, down, right, etc.
    path = pyastar2d.astar_path(weights, (0, 0), (4, 4), allow_diagonal=False)
    expected = np.array([[0, 0], [1, 0], [1, 1], [2, 1], [2, 2], [3, 2], [3, 3], [4, 3], [4, 4]])

    assert np.all(path == expected)


def test_no_solution():
    # Vertical wall.
    weights = np.ones((5, 5), dtype=np.float32)
    weights[:, 2] = np.inf

    path = pyastar2d.astar_path(weights, (0, 0), (4, 4), allow_diagonal=True)
    assert not path

    # Horizontal wall.
    weights = np.ones((5, 5), dtype=np.float32)
    weights[2, :] = np.inf

    path = pyastar2d.astar_path(weights, (0, 0), (4, 4), allow_diagonal=True)
    assert not path


def test_match_reverse():
    # Might fail if there are multiple paths, but this should be rare.
    h, w = 25, 25
    weights = (1.0 + 5.0 * np.random.random((h, w))).astype(np.float32)

    fwd = pyastar2d.astar_path(weights, (0, 0), (h - 1, w - 1))
    rev = pyastar2d.astar_path(weights, (h - 1, w - 1), (0, 0))

    assert np.all(fwd[::-1] == rev)

    fwd = pyastar2d.astar_path(weights, (0, 0), (h - 1, w - 1), allow_diagonal=True)
    rev = pyastar2d.astar_path(weights, (h - 1, w - 1), (0, 0), allow_diagonal=True)

    assert np.all(fwd[::-1] == rev)


def test_narrow():
    # Column weights.
    weights = np.ones((5, 1), dtype=np.float32)
    path = pyastar2d.astar_path(weights, (0, 0), (4, 0))

    expected = np.array([[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]])

    assert np.all(path == expected)

    # Row weights.
    weights = np.ones((1, 5), dtype=np.float32)
    path = pyastar2d.astar_path(weights, (0, 0), (0, 4))

    expected = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]])

    assert np.all(path == expected)


def test_bad_heuristic():
    # For valid heuristics, the cost to move must be at least 1.
    weights = (1.0 + 5.0 * np.random.random((10, 10))).astype(np.float32)
    # An element smaller than 1 should raise a ValueError.
    bad_cost = np.random.random() / 2.0
    weights[4, 4] = bad_cost

    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (9, 9))
        assert '.f' % bad_cost in exc.value.args[0]


def test_invalid_start_and_goal():
    weights = (1.0 + 5.0 * np.random.random((10, 10))).astype(np.float32)
    # Test bad start indices.
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (-1, 0), (9, 9))
        assert '-1' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (10, 0), (9, 9))
        assert '10' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, -1), (9, 9))
        assert '-1' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 10), (9, 9))
        assert '10' in exc.value.args[0]
    # Test bad goal indices.
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (-1, 9))
        assert '-1' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (10, 9))
        assert '10' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (0, -1))
        assert '-1' in exc.value.args[0]
    with pytest.raises(ValueError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (0, 10))
        assert '10' in exc.value.args[0]


def test_bad_weights_dtype():
    weights = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]], dtype=np.float64)
    with pytest.raises(AssertionError) as exc:
        pyastar2d.astar_path(weights, (0, 0), (2, 2))
    assert 'float64' in exc.value.args[0]


def test_orthogonal_x():
    weights = np.ones((5, 5), dtype=np.float32)
    path = pyastar2d.astar_path(
        weights,
        (0, 0),
        (4, 4),
        allow_diagonal=False,
        heuristic_override=Heuristic.ORTHOGONAL_X,
    )
    expected = np.array([[0, 0], [1, 0], [2, 0], [2, 1], [2, 2], [2, 3], [2, 4], [3, 4], [4, 4]])

    assert np.all(path == expected)


def test_orthogonal_y():
    weights = np.ones((5, 5), dtype=np.float32)
    path = pyastar2d.astar_path(
        weights,
        (0, 0),
        (4, 4),
        allow_diagonal=False,
        heuristic_override=Heuristic.ORTHOGONAL_Y,
    )
    expected = np.array([[0, 0], [0, 1], [0, 2], [1, 2], [2, 2], [3, 2], [4, 2], [4, 3], [4, 4]])

    assert np.all(path == expected)



import numpy as np

import pyastar2d

# The start and goal coordinates are in matrix coordinates (i, j).
start = (0, 0)
goal = (4, 4)

# The minimum cost must be 1 for the heuristic to be valid.
weights = np.array(
    [
        [1, 3, 3, 3, 3],
        [2, 1, 3, 3, 3],
        [2, 2, 1, 3, 3],
        [2, 2, 2, 1, 3],
        [2, 2, 2, 2, 1],
    ],
    dtype=np.float32,
)
print('Cost matrix:')
print(weights)
path = pyastar2d.astar_path(weights, start, goal, allow_diagonal=True)

# The path is returned as a numpy array of (i, j) coordinates.
print(f'Shortest path from {start} to {goal} found:')
print(path)



import ctypes
from enum import IntEnum
from typing import Optional, Tuple

import numpy as np

import pyastar2d.astar

# Define array types
ndmat_f_type = np.ctypeslib.ndpointer(dtype=np.float32, ndim=1, flags='C_CONTIGUOUS')
ndmat_i2_type = np.ctypeslib.ndpointer(dtype=np.int32, ndim=2, flags='C_CONTIGUOUS')

# Define input/output types
pyastar2d.astar.restype = ndmat_i2_type  # Nx2 (i, j) coordinates or None
pyastar2d.astar.argtypes = [
    ndmat_f_type,  # weights
    ctypes.c_int,  # height
    ctypes.c_int,  # width
    ctypes.c_int,  # start index in flattened grid
    ctypes.c_int,  # goal index in flattened grid
    ctypes.c_bool,  # allow diagonal
    ctypes.c_int,  # heuristic_override
]


class Heuristic(IntEnum):
    """The supported heuristics."""

    DEFAULT = 0
    ORTHOGONAL_X = 1
    ORTHOGONAL_Y = 2


def astar_path(
    weights: np.ndarray,
    start: Tuple[int, int],
    goal: Tuple[int, int],
    allow_diagonal: bool = False,
    heuristic_override: Heuristic = Heuristic.DEFAULT,
) -> Optional[np.ndarray]:
    """
    Run astar algorithm on 2d weights.

    param np.ndarray weights: A grid of weights e.g. np.ones((10, 10), dtype=np.float32)
    param Tuple[int, int] start: (i, j)
    param Tuple[int, int] goal: (i, j)
    param bool allow_diagonal: Whether to allow diagonal moves
    param Heuristic heuristic_override: Override heuristic, see Heuristic(IntEnum)

    """
    assert (
        weights.dtype == np.float32
    ), f'weights must have np.float32 data type, but has {weights.dtype}'
    # For the heuristic to be valid, each move must cost at least 1.
    if weights.min(axis=None) < 1.0:
        raise ValueError('Minimum cost to move must be 1, but got %f' % (weights.min(axis=None)))
    # Ensure start is within bounds.
    if start[0] < 0 or start[0] >= weights.shape[0] or start[1] < 0 or start[1] >= weights.shape[1]:
        raise ValueError(f'Start of {start} lies outside grid.')
    # Ensure goal is within bounds.
    if goal[0] < 0 or goal[0] >= weights.shape[0] or goal[1] < 0 or goal[1] >= weights.shape[1]:
        raise ValueError(f'Goal of {goal} lies outside grid.')

    height, width = weights.shape
    start_idx = np.ravel_multi_index(start, (height, width))
    goal_idx = np.ravel_multi_index(goal, (height, width))

    path = pyastar2d.astar.astar(
        weights.flatten(),
        height,
        width,
        start_idx,
        goal_idx,
        allow_diagonal,
        int(heuristic_override),
    )
    return path



import pathlib

from setuptools import Extension, find_packages, setup

# Use pathlib for paths
here = pathlib.Path(__file__).parent.resolve()

# Read README and requirements
long_description = (here / 'README.md').read_text(encoding='utf-8')
install_requires = (here / 'requirements.txt').read_text().splitlines()


class get_numpy_include:
    """Defer numpy import until it is actually installed."""

    def __str__(self):
        import numpy

        return numpy.get_include()


# Define the C++ extension
astar_module = Extension(
    name='pyastar2d.astar',
    sources=[
        'src/cpp/astar.cpp',
        'src/cpp/experimental_heuristics.cpp',
    ],
    define_macros=[
        ('Py_LIMITED_API', '0x03090000'),
        ('NPY_NO_DEPRECATED_API', 'NPY_2_0_API_VERSION'),
        ('NPY_TARGET_VERSION', 'NPY_2_0_API_VERSION'),
    ],
    py_limited_api=True,
    include_dirs=[
        'src/cpp',
        get_numpy_include(),
    ],
    extra_compile_args=['-O3', '-fpic', '-Wall'],
    language='c++',
)

# Define package metadata
setup(
    name='pyastar2d',
    use_scm_version=True,
    author='Hendrik Weideman',
    author_email='hjweide@gmail.com',
    description='A simple implementation of the A* algorithm for path-finding on a two-dimensional grid.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hjweide/pyastar2d',
    packages=find_packages(where='src', exclude=('tests',)),
    package_dir={'': 'src'},
    install_requires=install_requires,
    python_requires='>=3.9',
    ext_modules=[astar_module],
    options={'bdist_wheel': {'py_limited_api': 'cp39'}},
)
